/*
    cpufreqd

    Copyright (C) 2002 George Staikos <staikos@0wned.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#define _CPUFREQD_VERSION_ "1.0-pre1"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <getopt.h>
#include <unistd.h>
#include <libgen.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>



typedef struct power_info {
	char version[101];
	char time[101];
	int batteryPercent;
	int batteryTime;
	int ac;
	int flags;
} PowerInfo;

typedef struct cpu_policy {
	int lowerCpu;
	int upperCpu;
	int lowerPower;
	int upperPower;
	int isFreq:1;
	int isTime:1;
	int onAC:1;
	int powersave:1;
	struct cpu_policy *next;
} CpuPolicy;


CpuPolicy *newPolicy(void) {
CpuPolicy *p = malloc(sizeof(CpuPolicy));
	memset(p, 0, sizeof(CpuPolicy));
return p;
}


static CpuPolicy *policies = NULL;
static CpuPolicy *policyOn = NULL;
static int pollperiod = 10;



void printVersion(void) {
	printf("cpufreqd version %s.\n", _CPUFREQD_VERSION_);
	printf("Copyright 2002, George Staikos <staikos@0wned.org>\n");
}


void setPolicy(const CpuPolicy *p) {
FILE *fp;
int rc;
struct stat sb;
char sep = '%';

	if (!p)
		return;

	rc = stat("/proc/cpufreq", &sb);
	if (rc != 0) {
		perror("/proc/cpufreq");
		syslog(LOG_ERR, "Sorry, /proc/cpufreq is not accessible.");
		exit(-1);
	}
	
	fp = fopen("/proc/cpufreq", "w");
	if (!fp) {
		perror("/proc/cpufreq");
		syslog(LOG_ERR, "Sorry, /proc/cpufreq is not accessible.");
		exit(-1);
	}

	if (p->isTime)
		sep = ':';

	syslog(LOG_NOTICE, "New frequency scaling policy set: %d->%d, %s",
					p->lowerCpu, p->upperCpu,
					p->powersave ? "powersave"
							: "performance");

	/* FIXME: multiple cpu support */
	fprintf(fp, "0%c%d%c%d%c%s", sep, p->lowerCpu, sep, p->upperCpu,
					sep, p->powersave ? "powersave"
							: "performance");
	fclose(fp);
}


int readAPM(PowerInfo *dest) {
FILE *fp;
char buf[101];
char ignore3[101];
int ignore;
unsigned int ignore2;

	if (!dest)
		return 0;

	memset(dest, 0, sizeof(PowerInfo));

	fp = fopen("/proc/apm", "r");
	if (!fp) {
		return -1;
	}

	if (!fgets(buf, 100, fp)) {
		fclose(fp);
		return -1;
	}

	sscanf(buf, "%s %d.%d %x %x %x %x %d%% %d %s\n",
			dest->version, &ignore, &ignore,
			&dest->flags, &dest->ac, &ignore2,
			&ignore2, &dest->batteryPercent,
			&dest->batteryTime, ignore3);

	if (!strncmp(ignore3, "sec", 3)) {
		dest->batteryTime /= 60;
	}

	if (dest->batteryPercent > 100) {
		dest->batteryPercent = -1;
	}

	fclose(fp);

	return 0;
}






void mainLoop(void) {
PowerInfo p;
int rc;
int first = 1;

	for (;;) {

		if (!first)
			sleep(pollperiod);
		else first = 0;
		
		rc = readAPM(&p);
		if (rc < 0) {
			syslog(LOG_ERR, "Couldn't read power data from /proc/apm.");
			break;
		}

		if (policies) {
			CpuPolicy *c = policies;
			int index = 0;
			while (c) {
				int acCheck = ((c->onAC && p.ac) ||
					       (!c->onAC && !p.ac));

				int powerCheck;

				index++;
				if (c->isTime) {
					/* Battery Time is not supported? */
					if (p.batteryTime == 0 && !p.ac) {
						syslog(LOG_ERR, "Your kernel doesn't support battery time remaining, it seems.  Skipping rule %d.", index);
						continue;
					}

					powerCheck = ((c->lowerPower <= p.batteryTime) &&
							(c->upperPower >= p.batteryTime));
				} else {
					powerCheck = ((c->lowerPower <= p.batteryPercent) &&
							(c->upperPower >= p.batteryPercent));
				}

				if (acCheck && powerCheck) {
					if (policyOn != c) {
						policyOn = c;
						setPolicy(c);
					}
					break;
				}

				c = c->next;
			}
		}
	}
}






void daemonize(void) {
pid_t child;

	child = fork();
	if (child == 0) {
		mainLoop();
		closelog();
	} else {
		printf("Backgrounded.  pid=%u\n", child);
	}
}



char *stripWhiteSpace(char *str) {
int l = strlen(str);
char *newstr = str;

	while (isspace(str[l-1])) {
		str[l-1] = 0;
	}

	while (isspace(newstr[0])) {
		newstr++;
	}

return newstr;
}



void loadConfigFile(const char *filename) {
FILE *fp;
CpuPolicy *rec = 0;
int pstate = 0;
int line = 0;
int state = 0;
#define STATE_HAVE_MINPOWER   1
#define STATE_HAVE_MAXPOWER   2
#define STATE_HAVE_MINCPU     4
#define STATE_HAVE_MAXCPU     8
#define STATE_END            16

	assert(!policies);

	fp = fopen(filename, "r");
	if (!fp) {
		perror(filename);
		exit(-1);
	}

	while (!feof(fp)) {
		char buf[256];
		char *clean;

		if (!fgets(buf, 255, fp))
			break;

		buf[255] = 0;
		line++;

		clean = stripWhiteSpace(buf);

		if (clean[0] == 0) {
			continue;
		} else if (clean[0] == '#') {
			continue;
		} else if (clean[0] == '[') {
			if (!policies) {
				policies = newPolicy();
				rec = policies;
			} else if (rec) {
				if (state != STATE_END-1) {
					fprintf(stderr, "%s:%d: You must define cpu and frequency ranges in each rule.\n", filename, line);
					exit(-1);
				}
				rec->next = newPolicy();
				rec = rec->next;
			} else {
				assert(0);
			}
			pstate = 0;
			state = 0;
		} else if (!rec) {
			fprintf(stderr, "%s:%d: You must open a record first.\n", filename, line);
			exit(-1);
		} else {
			char *value = 0;
			char *name = strtok(clean, "=");
			if (!name) {
				fprintf(stderr, "%s:%d: Syntax error.\n", filename, line);
				exit(-1);
			}

			value = strtok(NULL, "");
			if (!value) {
				fprintf(stderr, "%s:%d: Value is missing.\n", filename, line);
				exit(-1);
			}

			name = stripWhiteSpace(name);
			value = stripWhiteSpace(value);

			if (strncasecmp(name, "minpower", 8) == 0) {
				if (state & STATE_HAVE_MINPOWER) {
					fprintf(stderr, "%s:%d: minpower redefined.\n", filename, line);
					exit(-1);
				}
				state |= STATE_HAVE_MINPOWER;
				rec->lowerPower = atoi(value);
			} else if (strncasecmp(name, "maxpower", 8) == 0) {
				if (state & STATE_HAVE_MAXPOWER) {
					fprintf(stderr, "%s:%d: maxpower redefined.\n", filename, line);
					exit(-1);
				}
				state |= STATE_HAVE_MAXPOWER;
				rec->upperPower = atoi(value);
			} else if (strncasecmp(name, "maxfreq", 7) == 0) {
				if (state & STATE_HAVE_MAXCPU) {
					fprintf(stderr, "%s:%d: maxfreq redefined.\n", filename, line);
					exit(-1);
				}
				state |= STATE_HAVE_MAXCPU;
				if (value[strlen(value)-1] == '%') {
					if (pstate == 0) {
						pstate = 1;
						rec->isFreq = 1;
					}

					if (pstate == 2) {
						fprintf(stderr, "%s:%d: You can't mix percentage and frequency in a rule.\n", filename, line);
						exit(-1);
					}

					/* truncate the % */
					value[strlen(value)-1] = 0;
				} else if (pstate == 1) { /* we're already %? */
						fprintf(stderr, "%s:%d: You can't mix percentage and frequency in a rule.\n", filename, line);
						exit(-1);
				} else { /* stay as raw freq */
					pstate = 2;
				}

				rec->upperCpu = atoi(value);
			} else if (strncasecmp(name, "minfreq", 7) == 0) {
				if (state & STATE_HAVE_MINCPU) {
					fprintf(stderr, "%s:%d: minfreq redefined.\n", filename, line);
					exit(-1);
				}
				state |= STATE_HAVE_MINCPU;
				if (value[strlen(value)-1] == '%') {
					if (pstate == 0) {
						pstate = 1;
						rec->isFreq = 1;
					}

					if (pstate == 2) {
						fprintf(stderr, "%s:%d: You can't mix percentage and frequency in a rule.\n", filename, line);
						exit(-1);
					}

					/* truncate the % */
					value[strlen(value)-1] = 0;
				} else if (pstate == 1) { /* we're already %? */
						fprintf(stderr, "%s:%d: You can't mix percentage and frequency in a rule.\n", filename, line);
						exit(-1);
				} else { /* stay as raw freq */
					pstate = 2;
				}

				rec->lowerCpu = atoi(value);
			} else if (strncasecmp(name, "ac", 2) == 0) {
				rec->onAC = !strncmp(value, "on", 2);
			} else if (strncasecmp(name, "profile", 7) == 0) {
				rec->powersave = !strncmp(value, "powersave", 9);
			} else {
				fprintf(stderr, "%s:%d: Unknown value.\n", filename, line);
				exit(-1);
			}
		}
	}

	fclose(fp);
}



int main(int argc, char **argv) {
int daemon = 0;
char filename[512] = "/etc/cpufreqd.conf";

	for (;;) {
		int option_index = 0;
		int c;
		static struct option long_options[] = {
			{"daemon", 0, 0, 'd'},
			{"version", 0, 0, 'v'},
			{"help", 0, 0, 'h'},
			{"poll", 1, 0, 'p'},
			{"file", 1, 0, 'f'},
			{0, 0, 0, 0}
		};

		c = getopt_long(argc, argv, "dp:f:h?",
				long_options, &option_index);
		if (c == -1)
			break;

		switch (c) {
		case 'd':
			daemon = 1;
		break;
		case 'p':
			pollperiod = atoi(optarg);
		break;
		case '?':
		case 'h':
			printVersion();
			printf("\nusage: %s <options>\n\n", argv[0]);
			printf("  -d, --daemon\t\t\tRun as a background daemon.\n");
			printf("  -p, --poll=seconds\t\tSet the poll interval in seconds.\n");
			printf("  -f, --file=filename\t\tSet the configuration filename.\n");
			printf("  -h, --help\t\t\tPrint help.\n");
			printf("  -v, --version\t\t\tPrint version information.\n");
			exit(0);
		case 'f':
			strncpy(filename, optarg, 500);
			filename[500] = 0;
		break;
		case 'v':
			printVersion();
			exit(0);
		default:
		break;
		}
	}

	loadConfigFile(filename);

	openlog(basename(argv[0]), LOG_PID, LOG_DAEMON);

	if (daemon) {
		daemonize();
	} else {
		mainLoop();
		closelog();
	}

return 0;
}

